import time
import os
import sys
import argparse
import json
import shutil

import numpy as np


class ImageClassification:

    def generate_class_dataset(self, datasource_loc, dst, ratio=0.8):
        origin_dir = os.path.join(datasource_loc, 'origin')
        dst_path = os.path.join(dst, 'standard', 'dataset_dir')
        train_dir = os.path.join(dst_path, 'training')
        os.makedirs(train_dir)
        val_dir = os.path.join(dst_path, 'validation')
        os.makedirs(val_dir)
        with open(os.path.join(origin_dir, 'label.json'), 'r') as rf:
            label_dict = json.load(rf)
        train_val_context = []
        for label in label_dict:
            if label_dict[label] == 0:
                continue
            label_files = os.listdir(
                os.path.join(datasource_loc, 'origin', label))
            train_val_context.extend(
                ["./imgs/" + label + "/" + label_file + " " + label + '\n'
                 for label_file in label_files]
            )
        train_val_context = np.array(train_val_context)
        np.random.shuffle(train_val_context)
        train_context, val_context = np.split(
            train_val_context, [int(ratio * len(train_val_context))])

        now_path = os.getcwd()
        with open(os.path.join(train_dir, 'training.csv'), 'w') as f:
            os.chdir(os.path.join(train_dir))
            for context in train_context:
                f.write(context)
                self.copy_img(context, origin_dir)
            os.chdir(now_path)

        with open(os.path.join(val_dir, 'validation.csv'), 'w') as f:
            os.chdir(os.path.join(val_dir))
            for context in val_context:
                f.write(context)
                self.copy_img(context, origin_dir)
            os.chdir(now_path)

    def copy_img(self, lable_content, src_path):
        dst_img_path, lable = lable_content.split(' ')
        src_img_path = os.path.join(src_path, lable.strip(),
                                    os.path.basename(dst_img_path))
        if not os.path.exists(os.path.dirname(dst_img_path)):
            os.makedirs(os.path.dirname(dst_img_path))
        shutil.copy(src_img_path, dst_img_path)


class ObjectDetection:
    def __init__(self, datasource_loc, dst, ratio=0.8):
        self.datasource_loc = datasource_loc
        self.dst = dst
        self.ratio = ratio

    def generate_object_all_dataset(self):
        if os.path.exists(os.path.join(self.dst, "dataset_dir")):
            os.rmdir(os.path.join(self.dst, "dataset_dir"))
        dst = self.dst
        self.dst = os.path.join(dst, "standard")
        self.generate_object_standard_dataset()
        print("[Success]:The default dataset format has been migrated.")
        self.dst = os.path.join(dst, "autodl")
        self.generate_object_autodl_dataset()
        print("[Success]:The autodl dataset format has been migrated.")

    def generate_object_standard_dataset(self):
        for name in ("Annotations", "JPEGImages", "category.txt"):
            src_dir = os.path.join(self.datasource_loc, "origin", name)
            if os.path.isdir(src_dir):
                shutil.copytree(src_dir,
                                os.path.join(self.dst, "dataset_dir", name))
            else:
                shutil.copy(src_dir,
                            os.path.join(self.dst, "dataset_dir", name))
        annotations_dir = os.path.join(self.datasource_loc,
                                       "origin/Annotations")
        train_val_files = os.listdir(annotations_dir)
        train_val_names = []
        for file_name in train_val_files:
            file_id, ext = os.path.splitext(file_name)
            if ext == ".xml":
                train_val_names.append(file_id)
        train_val_names = np.array(train_val_names)
        np.random.shuffle(train_val_names)
        train_names, val_names = np.split(
            train_val_names, [int(self.ratio * len(train_val_names))])
        train_val_dir = os.path.join(
            self.dst, "dataset_dir", 'ImageSets/Main')
        os.makedirs(train_val_dir)
        with open(
                os.path.join(train_val_dir, 'train.txt'), 'w') as f:
            for name in train_names:
                f.write(name + '\n')
        with open(
                os.path.join(train_val_dir, 'val.txt'), 'w') as f:
            for name in val_names:
                f.write(name + '\n')
        # Migrate unlabeled images
        unlabel_path = os.path.join(self.datasource_loc, "unlabel")
        unlabel_images = os.listdir(unlabel_path)
        if not unlabel_images:
            return
        for image in unlabel_images:
            shutil.copy(
                os.path.join(unlabel_path, image),
                os.path.join(self.dst, "dataset_dir", "JPEGImages")
            )

    def generate_object_autodl_dataset(self):
        # datasource content directory
        origin_dir = os.path.join(self.datasource_loc, 'origin')
        annotations_dir = os.path.join(origin_dir, 'Annotations')
        images_dir = os.path.join(origin_dir, 'JPEGImages')
        # get labeled images and split images to train and val by ratio
        train_val_files = os.listdir(annotations_dir)
        train_val_names = []
        for file_name in train_val_files:
            file_id, ext = os.path.splitext(file_name)
            if ext == '.xml':
                train_val_names.append(file_id)
        train_val_names = np.array(train_val_names)
        np.random.shuffle(train_val_names)
        train_names, val_names = np.split(
            train_val_names, [round(self.ratio * len(train_val_names))])
        # new dir: train and val
        train_dir = os.path.join(self.dst, "dataset_dir", "train")
        train_dir_jpg = os.path.join(self.dst, "dataset_dir", "train/images")
        val_dir = os.path.join(self.dst, "dataset_dir", "val")
        val_dir_jpg = os.path.join(self.dst, "dataset_dir", "val/images")
        os.makedirs(train_dir_jpg)
        os.makedirs(val_dir_jpg)
        # open and write train/Annotations.zip
        result_train = self.make_zip_by_files(
            os.path.join(train_dir, 'Annotations.zip'), {
                os.path.join(
                    annotations_dir, image + '.xml'): image + '.xml'
                for image in train_names
            })
        self.check_result_of_create_zip_file(result_train)

        # open and write val/Annotations.zip
        result_val = self.make_zip_by_files(
            os.path.join(val_dir, 'Annotations.zip'), {
                os.path.join(
                    annotations_dir, image + '.xml'): image + '.xml'
                for image in val_names
            })
        self.check_result_of_create_zip_file(result_val)

        # copy train jpg images to train/images
        for image in train_names:
            shutil.copy(os.path.join(images_dir, image + ".jpg"),
                        os.path.join(train_dir_jpg, image + ".jpg"))

        # copy val jpg images to val/images
        for image in val_names:
            shutil.copy(os.path.join(images_dir, image + ".jpg"),
                        os.path.join(val_dir_jpg, image + ".jpg"))

        # copy unlabeled images
        unlabel_path = os.path.join(self.datasource_loc, "unlabel")
        unlabel_images = os.listdir(unlabel_path)
        if not unlabel_images:
            return
        for image in unlabel_images:
            shutil.copy(
                os.path.join(unlabel_path, image),
                os.path.join(train_dir_jpg)
            )

    @staticmethod
    def check_result_of_create_zip_file(result):
        if not isinstance(result, dict) or 'created' not in result:
            raise "Task datasource generate error."
        if not result['created']:
            raise "Task datasource generate error.%s" % (result['reason'])

    def make_zip_by_files(self, filename, files_dict):
        """
        :param filename: type string
          - new zip file abspath with name
        :param files_dict: type: dict
          - key, compression file abspath
          - value, path in zip file
        :return:
        """
        import zipfile
        try:
            with zipfile.ZipFile(filename, 'w') as zp:
                for key in files_dict:
                    zp.write(key, files_dict[key])
        except Exception as e:
            return dict(created=False,
                        reason=f'Write zip file error: {e}')
        return dict(created=True, reason='')


class InstanceSegmentation:
    def __init__(self, datasource_loc, dst, ratio=0.8):
        self.datasource_loc = datasource_loc
        self.dst = os.path.join(dst, "standard", "dataset_dir")
        self.ratio = ratio
        self.image_id = 1
        self.annotation_id = 1

    def generate_instance_dataset(self):
        origin_annotation = os.path.join(
            self.datasource_loc, "origin", "Annotations"
        )
        origin_images = os.path.join(
            self.datasource_loc, "origin", "JPEGImages"
        )
        category_file = os.path.join(
            self.datasource_loc, "origin", "category.json"
        )
        dst_annotations = os.path.join(self.dst, "annotations")
        os.makedirs(dst_annotations)
        dst_train = os.path.join(self.dst, "train")
        os.makedirs(dst_train)
        dst_val = os.path.join(self.dst, "val")
        os.makedirs(dst_val)
        # generate comments and copy comment images based on comment information
        file_list = os.listdir(origin_annotation)
        train_val_files = [ann_file for _, ann_file in enumerate(file_list)]
        train_val_files = np.array(train_val_files)
        np.random.shuffle(train_val_files)
        train_files, val_files = np.split(
            train_val_files, [int(self.ratio * len(train_val_files))])
        with open(category_file, 'r') as f:
            category_json = json.load(f)
        instances_train = self._generate_annotations_data(
            origin_annotation, origin_images,
            category_json, train_files, dst_train)
        with open(os.path.join(dst_annotations, "instances_train.json"),
                  'w') as f:
            json.dump(instances_train, f)
        instances_val = self._generate_annotations_data(
            origin_annotation, origin_images,
            category_json, val_files, dst_val)
        with open(os.path.join(dst_annotations, "instances_val.json"),
                  'w') as f:
            json.dump(instances_val, f)
        # Migrate unlabeled images
        unlabel_path = os.path.join(self.datasource_loc, "unlabel")
        unlabel_images = os.listdir(unlabel_path)
        if not unlabel_images:
            return
        for image in unlabel_images:
            shutil.copy(
                os.path.join(unlabel_path, image),
                os.path.join(dst_train)
            )

    def _generate_annotations_data(self, origin_annotation, origin_images,
                                   category_json, file_list, dst_image):
        annotations_data = category_json
        annotations_data.update({"images": [], "annotations": []})
        for file in file_list:
            anno_file = os.path.join(origin_annotation, file)
            with open(anno_file, 'r') as f:
                train_json = json.load(f)
            # generate annotations
            image_dict = train_json["images"][0]
            image_dict["id"] = self.image_id
            annotations_data["images"].append(image_dict)
            annotations_dict = train_json["annotations"]
            for annotation in annotations_dict:
                annotation["id"] = self.annotation_id
                annotation["image_id"] = self.image_id
                annotations_data["annotations"].append(annotation)
                self.annotation_id += 1
            self.image_id += 1
            image_path = os.path.join(
                origin_images, os.path.splitext(file)[0] + ".jpg")
            shutil.copy(image_path, dst_image)
        return annotations_data


class TextClassification:

    def generate_text_dataset(self, datasource_loc, dst, ratio=0.8):
        origin_dir = os.path.join(datasource_loc, 'origin')
        file_list = os.listdir(origin_dir)
        train_val_context = []
        for file in file_list:
            _, ext = os.path.splitext(file)
            if ext != ".tsv":
                continue
            with open(os.path.join(origin_dir, file), 'r') as rf:
                for line in rf:
                    train_val_context.append(line)
        train_val_context = np.array(train_val_context)
        np.random.shuffle(train_val_context)
        train_context, val_context = np.split(
            train_val_context, [int(ratio * len(train_val_context))])

        dst_path = os.path.join(dst, "dataset_dir")
        if not os.path.exists(dst_path):
            os.makedirs(dst_path)
        with open(os.path.join(dst_path, 'training.tsv'), 'w') as f:
            for context in train_context:
                f.write(context)
        with open(os.path.join(dst_path, 'validation.tsv'), 'w') as f:
            for context in val_context:
                f.write(context)


class BaseMigrations:
    def __init__(self, args, datasource):
        self.args = args
        self.datasource = datasource
        self.dataset_dirs = {
            "Image Classification": 1,
            "Object Detection": 1,
            "Instance Segmentation": 1,
            "Text Classification": 1
        }

    def get_dataset_list(self):
        u_uuids = os.listdir(self.datasource)
        if not u_uuids:
            print("There is no dataset to be migrated.")
            sys.exit(0)
        dataset_info_list = []
        for u_uuid in u_uuids:
            dataset_info_list.append(self.get_dataset_info(u_uuid))
        return dataset_info_list

    def get_dataset_info(self, u_uuid):
        dataset_dir = os.path.join(self.datasource, u_uuid)
        stat = os.stat(dataset_dir)
        mtime = time.strftime(
            "%Y-%m-%d %H:%M:%S", time.localtime(stat.st_mtime))
        scenario = self.get_dataset_scenario(dataset_dir)
        sample_size = self.get_dataset_sample_and_size(dataset_dir, scenario)
        dataset_info = {"UUID": u_uuid, "Scenario": scenario}
        dataset_info.update(sample_size)
        dataset_info["Last Modified"] = mtime
        return dataset_info

    def get_dataset_scenario(self, dataset_dir):
        dirs = os.listdir(dataset_dir)
        if "unlabel" in dirs:
            # object and instance
            origin_dir = os.path.join(dataset_dir, "origin")
            if "ImageSets" in os.listdir(origin_dir):
                return "Object Detection"
            else:
                return "Instance Segmentation"
        else:
            # image and text
            if "thumbnail" not in dirs:
                return "Text Classification"
            else:
                return "Image Classification"

    def get_dataset_sample_and_size(self, dataset_dir, scenario):
        if scenario == "Image Classification":
            origin_dir = os.path.join(dataset_dir, "origin")
            record_file_dir = os.path.join(dataset_dir, "thumbnail")
            total_num = self._get_record_file_num(record_file_dir)
            size_b = self._get_dir_size_b(origin_dir)
        elif scenario == "Text Classification":
            origin_dir = os.path.join(dataset_dir, "origin")
            size_b = self._get_dir_size_b(origin_dir)
            total_num = 0
            origin_files = os.listdir(origin_dir)
            for file in origin_files:
                if file.endswith(".tsv"):
                    with open(os.path.join(origin_dir, file), 'r') as f:
                        for line in f:
                            total_num += 1
        else:
            unlabel_dir = os.path.join(dataset_dir, "unlabel")
            origin_dir = os.path.join(dataset_dir, "origin")
            record_file_path = os.path.join(
                dataset_dir,
                "record_file.txt"
            )
            total_num = self._get_record_file_num(record_file_path)
            size_b = self._get_dir_size_b(unlabel_dir)
            size_b += self._get_dir_size_b(origin_dir)
        size = round(size_b / 1024 / 1024, 2)
        return {"Samples": total_num, "Size(MB)": size}

    def _get_record_file_num(self, record_file_path):
        def count_record_file(path):
            with open(path, "r") as f:
                return len(f.readlines())

        if os.path.isdir(record_file_path):
            count = 0
            names = os.listdir(record_file_path)
            for name in names:
                if name.endswith("_record_file.txt"):
                    path = os.path.join(record_file_path, name)
                    num = count_record_file(path)
                    count += num
            return count
        elif os.path.isfile(record_file_path):
            return count_record_file(record_file_path)
        else:
            return 0

    def _get_dir_size_b(self, path):
        if os.path.isdir(path):
            return sum([sum([os.path.getsize(os.path.join(
                root, name)) for name in files])
                for root, dirs, files in os.walk(path)])
        else:
            return 0

    def migrate_dataset(self, u_uuid, scenario=None):
        dataset_dir = os.path.join(self.datasource, u_uuid)
        if not scenario:
            scenario = self.get_dataset_scenario(dataset_dir)
        output_dir = self._check_and_create_dataset_dir(scenario)
        if scenario == "Image Classification":
            image_obj = ImageClassification()
            image_obj.generate_class_dataset(dataset_dir, output_dir)
        elif scenario == "Object Detection":
            obj_detection = ObjectDetection(dataset_dir, output_dir)
            if self.args.type_object == "standard":
                obj_detection.generate_object_standard_dataset()
            elif self.args.type_object == "autodl":
                obj_detection.generate_object_autodl_dataset()
            else:
                obj_detection.generate_object_all_dataset()
        elif scenario == "Instance Segmentation":
            instance = InstanceSegmentation(dataset_dir, output_dir)
            instance.generate_instance_dataset()
        elif scenario == "Text Classification":
            text_obj = TextClassification()
            text_obj.generate_text_dataset(dataset_dir, output_dir)
        print("[Tips]:The migration of %s dataset is complete." % scenario)
        if self.args.delete:
            shutil.rmtree(dataset_dir)
            print("[Delete]: Deleted directory %s." % dataset_dir)
        return output_dir

    def migrate_all_dataset(self):
        self._check_and_create_output_dir()
        dataset_info_list = self.get_dataset_list()
        for dataset_info in dataset_info_list:
            dst_dir = self.migrate_dataset(
                dataset_info["UUID"], dataset_info["Scenario"]
            )
            dataset_info["Output Dir"] = os.path.basename(dst_dir)
        print("\033[1;33mTips: The output base directory of the dataset is "
              "%s.\033[0m" % os.path.abspath(self.args.output_dir))
        format_print(dataset_info_list)

    def _check_and_create_output_dir(self):
        self.args.output_dir = os.path.join(
            self.args.output_dir, "datasets_from_aistudio")
        if os.path.exists(self.args.output_dir):
            n = 1
            while 1:
                new_output_dir = self.args.output_dir + "_%s" % n
                if not os.path.exists(new_output_dir):
                    self.args.output_dir = new_output_dir
                    break
                n += 1
        os.makedirs(self.args.output_dir)

    def _check_and_create_dataset_dir(self, scenario):
        if scenario == "Image Classification":
            dataset_dir = "image_classification"
        elif scenario == "Object Detection":
            dataset_dir = "object_detection"
        elif scenario == "Instance Segmentation":
            dataset_dir = "instance_segmentation"
        elif scenario == "Text Classification":
            dataset_dir = "text_classification"
        dst_dir = os.path.join(self.args.output_dir, dataset_dir)
        if os.path.exists(dst_dir):
            while 1:
                new_dst_dir = dst_dir + "_%s" % self.dataset_dirs.get(scenario)
                self.dataset_dirs[scenario] = self.dataset_dirs[scenario] + 1
                if not os.path.exists(new_dst_dir):
                    dst_dir = new_dst_dir
                    break
        os.makedirs(dst_dir)
        return dst_dir


def parse_args():
    parser = argparse.ArgumentParser(
        description="Migrate the AIStudio dataset")
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-a", dest="all", action="store_true",
                       help="Migrate all datasets")
    group.add_argument("-l", dest="list", action="store_true",
                       help="List AIStudio dataset information.")
    group.add_argument("-u", dest="u_uuid",
                       help="AIStudio dataset UUID.", type=str)
    parser.add_argument("-o", dest="output_dir",
                        help="The output directory after migration.",
                        type=str)
    parser.add_argument("-H", dest="home",
                        help="""An attempt is made to obtain the user"s home 
                        directory. If this fails, the user is required to 
                        pass in this parameter.""",
                        default=os.environ.get("HOME", ""))
    parser.add_argument("-t", dest="type_object",
                        help="Type of the object detection dataset to be "
                             "exported. standard: LaDL Default Format; "
                             "autodl: LaDL AutoDL Format; all[default]: Both "
                             "standard and autodl are generated.",
                        type=str, choices=["standard", "autodl", "all"])
    parser.add_argument("-d", dest="delete", action="store_true",
                        help="Delete the AIStudio dataset after migration.")
    return parser.parse_args()


def check_args(args, datasource):
    if not os.path.exists(args.home):
        print("[Error]:The user's home directory fails to be obtained. "
              "Please use the -H parameter to specify the home directory.")
        sys.exit(1)
    if not os.path.exists(datasource):
        print("There is no dataset to be migrated. Reason: "
              "The data source directory %s does not exist." % datasource)
        sys.exit(0)
    if args.u_uuid and not os.path.exists(
            os.path.join(datasource, args.u_uuid)
    ):
        print("[Error]:The UUID does not exist.")
        sys.exit(1)
    elif args.u_uuid and args.output_dir:
        if not os.path.exists(os.path.join(datasource, args.u_uuid)):
            print("[Error]:The UUID does not exist.")
            sys.exit(1)
        if not os.path.exists(args.output_dir):
            print("[Error]:The output dir does not exist.")
            sys.exit(1)
    if args.all and not args.output_dir:
        print("[Error]:Output directory must be specified.")
        sys.exit(1)
    elif args.all and not os.path.exists(args.output_dir):
        print("[Error]:The output directory does not exist.")
        sys.exit(1)


def format_print(data):
    # Extract table head
    keyHeader = data[0].keys()
    # Store the maximum length of each column
    keyMaxLen = {}

    for item in data:
        for i, h in enumerate(keyHeader):
            # Calculate the maximum length corresponding to each attribute
            maxLen = max(len(h), len(str(item[h])))
            if keyMaxLen.get(h, None):
                maxLen = max(maxLen, keyMaxLen[h])
            keyMaxLen[h] = maxLen

    def printGroup(group):
        for item in group:
            print('\r')
            for i, h in enumerate(keyHeader):
                itemLen = keyMaxLen.get(h, str(h)) + 4
                # fill space and center
                s = str(item[h]).center(
                    itemLen, '-' if item[h] == '-' else ' ')

                icon = '|'
                if item[h] == '-':
                    icon = '+'
                s = (icon if i == 0 else '') + s + icon
                print(s, end='')
        print('\r')

    print('\n[Datasets info]', end='')

    tag = {}
    header = {}
    for i, h in enumerate(keyHeader):
        tag[h] = '-'
        header[h] = h
    # add third row tag
    data.insert(0, tag)
    # add second row header
    data.insert(0, header)
    # add first row tag
    data.insert(0, tag)
    # add last row tag
    data.append(tag)

    printGroup(data)


def main():
    args = parse_args()
    datasource = os.path.join(args.home, "ai_studio", "datasource")
    check_args(args, datasource)
    migrate = BaseMigrations(args, datasource)
    print("\033[1;33mTips: The directory where AIStudio resides is "
          "%s.\033[0m" % datasource)
    if args.list:
        dataset_info_list = migrate.get_dataset_list()
        format_print(dataset_info_list)
    elif args.u_uuid:
        if not args.output_dir:
            dataset_info = migrate.get_dataset_info(args.u_uuid)
            format_print([dataset_info])
        else:
            print("Start migration ......")
            dst_dir = migrate.migrate_dataset(args.u_uuid)
            print("The export path is %s." % dst_dir)
            print("End migration !")
    elif args.all:
        migrate.migrate_all_dataset()


if __name__ == "__main__":
    main()
